/***************************************************************************
 * Copyright 2012 Kieker Project (http://kieker-monitoring.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.beans.application;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;

import javax.faces.application.FacesMessage;

import org.springframework.beans.factory.annotation.Autowired;

import kieker.analysis.model.analysisMetaModel.MIProject;
import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;
import kieker.webgui.beans.view.CurrentProjectOverviewBean;
import kieker.webgui.common.IProjectManagerFacade;
import kieker.webgui.common.exception.ProjectAlreadyExistingException;
import kieker.webgui.common.exception.ProjectLoadException;
import kieker.webgui.common.exception.ProjectNotExistingException;

/**
 * The {@link ProjectsBean} is a Spring managed bean to manage a list with all application wide available projects. It provides methods to receive this list as well
 * as methods to add, create, rename, open and copy projects. Furthermore the state of existing projects (like the timestamp or the state of the analysis) can be
 * received via this bean. In order to realize a good abstraction, this bean should be used for most accesses to the projects. The necessary synchronization is
 * achieved in the {@link IProjectManagerFacade}.<br>
 * As this bean contains the whole list of the available projects, it is session scoped. There is no reason for multiple instances of this class.
 * 
 * @author Nils Christian Ehmke
 */
public final class ProjectsBean {

	private static final Log LOG = LogFactory.getLog(ProjectsBean.class);
	private static final String DEFAULT_TIMESTAMP = "N/A";
	private static final String DEFAULT_ANALYSIS_STATE = "";

	private final List<String> projects = Collections.synchronizedList(new ArrayList<String>());

	@Autowired
	private IProjectManagerFacade projectManagerFacade;

	/**
	 * Default constructor. <b>Do not use this constructor. This bean is Spring managed.</b>
	 */
	public ProjectsBean() {
		// No code necessary
	}

	/**
	 * The Setter for the property {@link ProjectsBean#projectManagerFacade}. <b>Do not use this method. This property is Spring managed.</b>
	 * 
	 * @param projectManagerFacade
	 *            The new value for the property.
	 */
	public void setProjectManagerFacade(final IProjectManagerFacade projectManagerFacade) {
		this.projectManagerFacade = projectManagerFacade;
	}

	/**
	 * This method does some initialization work after construction. <b>Do not call this method. The method is Spring managed.</b>
	 */
	protected void initialize() {
		// Load a list with all available projects on the FS
		this.projects.addAll(this.projectManagerFacade.listAllProjects());
	}

	/**
	 * This method adds a new project with the given name to the application if this is possible. In either case (this means if the project has been created, if the
	 * project could not be created for any reason) the user will be informed via the growl component. In other words: A message will be delivered within the current
	 * context. The creation of a new project means in this context that a directory with its name will be created and that an empty but valid kax-file will be
	 * produced. After creating the project (provided that he creation was successful) the projects list of the "calling" {@link CurrentProjectOverviewBean} will be
	 * updated.<br>
	 * If something should go wrong, no exception will be thrown. Everything is caught if necessary.
	 * 
	 * @param project
	 *            The name for the new project which should be added to the application.
	 * @param currentProjectOverviewBean
	 *            The current instance of {@link CurrentProjectOverviewBean} which will be used to update the list of projects.
	 */
	public void addProject(final String project, final CurrentProjectOverviewBean currentProjectOverviewBean) {
		try {
			// Try and use the FS-Manager to create the project atomically.
			this.projectManagerFacade.addProject(project);
			// If there were no exception, everything went well. We can add the project to our list.
			this.projects.add(project);
			// Inform the user
			GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_INFO, "Project created.");
			// Update the list of the "calling" bean
			currentProjectOverviewBean.updateLists();
		} catch (final IOException ex) {
			ProjectsBean.LOG.error("An error occured while creating the project.", ex);
			GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occured while creating the project.");
		} catch (final ProjectAlreadyExistingException ex) {
			ProjectsBean.LOG.info("A project with the same name exists already.", ex);
			GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_WARN, "A project with the same name exists already.");
		}
	}

	/**
	 * This method tries to copy the project with the given name and saves it under the given new name. In either case (this means if the project has been created,
	 * if the project could not be created for any reason) the user will be informed via the growl component. In other words: A message will be delivered within the
	 * current context. After creating the project (provided that he creation was successful) the projects list of the "calling" {@link CurrentProjectOverviewBean}
	 * will be updated.<br>
	 * If something should go wrong, no exception will be thrown. Everything is caught if necessary.
	 * 
	 * @param sourceProject
	 *            The name of the source project.
	 * @param destinationProject
	 *            The name of the new (copied) project.
	 * @param currentProjectOverviewBean
	 *            The current instance of {@link CurrentProjectOverviewBean} which will be used to update the list of projects.
	 */
	public void copyProject(final String sourceProject, final String destinationProject, final CurrentProjectOverviewBean currentProjectOverviewBean) {
		try {
			// Try and use the FS-Manager to copy the project atomically.
			this.projectManagerFacade.copyProject(sourceProject, destinationProject);
			// If there were no exception, everything went well. We can add the project to our list.
			this.projects.add(destinationProject);
			// Inform the user
			GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_INFO, "Project created.");
			// Update the list of the "calling" bean
			currentProjectOverviewBean.updateLists();
		} catch (final IOException ex) {
			ProjectsBean.LOG.error("An error occured while creating the project.", ex);
			GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occured while creating the project.");
			ex.printStackTrace();
		} catch (final ProjectAlreadyExistingException ex) {
			ProjectsBean.LOG.info("A project with the same name exists already.", ex);
			GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_WARN, "A project with the same name exists already.");
		} catch (final ProjectNotExistingException ex) {
			ProjectsBean.LOG.info("A project with the given name does not exist.", ex);
			GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_WARN, "A project with the given name does not exist.");
		}
	}

	/**
	 * This method can be used to open an already existing project. This means that the current state of the project on the file system is loaded into an instance of
	 * {@link MIProject}. This instance can be modified at will and for example later saved by the {@link IProjectManagerFacade}.
	 * 
	 * @param project
	 *            The name of the project to be opened.
	 * @return Either the model instance if the loading was successful or null otherwise. In the latter case, the user will be informed via the growl-component.
	 * @throws ProjectLoadException
	 *             If something went wrong during the loading of the project.
	 */
	public MIProject openProject(final String project) throws ProjectLoadException {
		try {
			return this.projectManagerFacade.openProject(project);
		} catch (final IOException ex) {
			ProjectsBean.LOG.error("An error occured while loading the project.", ex);
			throw new ProjectLoadException("An error occured while loading the project.", ex);
		} catch (final ProjectNotExistingException ex) {
			ProjectsBean.LOG.info("A project with the given name does not exist.", ex);
			throw new ProjectLoadException("A project with the given name does not exist.", ex);
		}
	}

	/**
	 * This method can be used to get the current time stamp of a given project as a human readable date. If the project doesn't exist or the time stamp would be
	 * invalid "N/A" will be returned.
	 * 
	 * @param project
	 *            The project whose time stamp should be collected.
	 * @return The human readable time stamp of the project.
	 */
	public String getCurrTimeStamp(final String project) {
		try {
			// Get the current time stamp of the project
			final long timeStamp = this.projectManagerFacade.getCurrTimeStamp(project);
			// Convert the stamp into a human readable string.
			return new Date(timeStamp).toString();
		} catch (final ProjectNotExistingException ex) {
			ProjectsBean.LOG.info("A project with the given name does not exist.", ex);
			// We can assume that something went wrong
			return ProjectsBean.DEFAULT_TIMESTAMP;
		}

	}

	/**
	 * This method delivers all available projects as a list of string. The list itself is just a copy and can be modified as will. Keep in mind that this means also
	 * that every call to this method will create a new copy.
	 * 
	 * @return All currently available projects as a copied and modifiable list.
	 */
	public List<String> getProjects() {
		synchronized (this.projects) {
			return new ArrayList<String>(this.projects);
		}
	}

	/**
	 * This method can be used to deliver the state of the analysis controller of the given project as a human readable string.
	 * 
	 * @param project
	 *            The project whose state should be delivered.
	 * @return The current state of the corresponding AnalysisController.
	 */
	public String getAnalysisControllerState(final String project) {
		try {
			return this.projectManagerFacade.getCurrentState(project).toString();
		} catch (final ProjectNotExistingException ex) {
			ProjectsBean.LOG.info("A project with the given name does not exist.", ex);
			return ProjectsBean.DEFAULT_ANALYSIS_STATE;
		}
	}
}
