/***************************************************************************
 * Copyright 2012 Kieker Project (http://kieker-monitoring.net)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/

package kieker.webgui.beans.view;

import java.io.IOException;
import java.lang.annotation.Annotation;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.List;
import java.util.UUID;

import javax.faces.application.Application;
import javax.faces.application.FacesMessage;
import javax.faces.component.UIComponent;
import javax.faces.component.UIInput;
import javax.faces.component.html.HtmlOutputText;
import javax.faces.context.FacesContext;

import org.springframework.beans.factory.annotation.Autowired;

import kieker.analysis.display.annotation.Display;
import kieker.analysis.model.analysisMetaModel.MIAnalysisMetaModelFactory;
import kieker.analysis.model.analysisMetaModel.MIDisplay;
import kieker.analysis.model.analysisMetaModel.MIDisplayConnector;
import kieker.analysis.model.analysisMetaModel.MIProject;
import kieker.analysis.model.analysisMetaModel.MIView;
import kieker.analysis.model.analysisMetaModel.impl.MAnalysisMetaModelFactory;
import kieker.analysis.plugin.AbstractPlugin;
import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;
import kieker.webgui.beans.application.GlobalPropertiesBean;
import kieker.webgui.beans.application.ProjectsBean;
import kieker.webgui.common.ClassAndMethodContainer;
import kieker.webgui.common.IProjectManagerFacade;
import kieker.webgui.common.exception.NewerProjectException;
import kieker.webgui.common.exception.ProjectLoadException;
import kieker.webgui.common.exception.ProjectNotExistingException;

import org.primefaces.component.dashboard.Dashboard;
import org.primefaces.component.panel.Panel;
import org.primefaces.context.RequestContext;
import org.primefaces.event.TabChangeEvent;
import org.primefaces.model.DashboardColumn;
import org.primefaces.model.DashboardModel;
import org.primefaces.model.DefaultDashboardColumn;
import org.primefaces.model.DefaultDashboardModel;

/**
 * The {@link CurrentCockpitEditorBean} contains the necessary data behind an instance of the cockpit editor.
 * The class is a Spring managed bean with view scope to make sure that one user (even in one session) can open multiple projects at a time without causing any
 * problems.
 * 
 * @author Nils Christian Ehmke
 */
public final class CurrentCockpitEditorBean {

	private static final int NUMBER_COLUMNS = 2;
	private static final Log LOG = LogFactory.getLog(CurrentCockpitEditorBean.class);

	private final MIAnalysisMetaModelFactory factory = new MAnalysisMetaModelFactory();
	@Autowired
	private IProjectManagerFacade projectManagerFacade;
	private ClassAndMethodContainer classAndMethodContainer;

	private long timeStamp;
	private String projectName;
	private MIProject project;
	private MIView activeView;
	private ClassLoader classLoader;
	private Dashboard dashboard;
	private DashboardModel dashboardModel;

	@Autowired
	private ProjectsBean projectsBean;

	@Autowired
	private GlobalPropertiesBean globalPropertiesBean;

	/**
	 * Creates a new instance of this class. <b>Do not call this constructor manually. It will only be accessed by Spring.</b>
	 */
	public CurrentCockpitEditorBean() {
		this.createDashboard();
	}

	/**
	 * Setter for the property {@link CurrentCockpitEditorBean#projectManagerFacade}.
	 * 
	 * @param projectManagerFacade
	 *            The new value of the property.
	 */
	public void setProjectManagerFacade(final IProjectManagerFacade projectManagerFacade) {
		this.projectManagerFacade = projectManagerFacade;
	}

	private void createDashboard() {
		final FacesContext fc = FacesContext.getCurrentInstance();
		final Application application = fc.getApplication();

		// Create the primefaces dashboard
		this.dashboard = (Dashboard) application.createComponent(fc, "org.primefaces.component.Dashboard", "org.primefaces.component.DashboardRenderer");
		this.dashboard.setId("dashboard");

		// Create the model and add the columns
		this.dashboardModel = new DefaultDashboardModel();
		for (int i = 0; i < CurrentCockpitEditorBean.NUMBER_COLUMNS; i++) {
			final DashboardColumn column = new DefaultDashboardColumn();
			this.dashboardModel.addColumn(column);
		}
		this.dashboard.setModel(this.dashboardModel);
	}

	private void fillDashboard() {
		// Dump the old entries
		this.clearDashboard();

		// Now add the entries from the current view
		if (this.activeView != null) {
			final FacesContext fc = FacesContext.getCurrentInstance();
			final Application application = fc.getApplication();

			// Add a panel for every display connector we have
			final List<MIDisplayConnector> connectors = this.activeView.getDisplayConnectors();
			int i = 0;
			for (final MIDisplayConnector connector : connectors) {
				final Panel panel = (Panel) application.createComponent(fc, "org.primefaces.component.Panel", "org.primefaces.component.PanelRenderer");
				panel.setId("displayConnectorPanel_" + i);
				panel.setHeader(connector.getName());
				panel.setClosable(true);
				panel.setToggleable(false);

				this.getDashboard().getChildren().add(panel);
				final DashboardColumn column = this.dashboard.getModel().getColumn(i % CurrentCockpitEditorBean.NUMBER_COLUMNS);
				column.addWidget(panel.getId());
				final HtmlOutputText text = new HtmlOutputText();
				text.setValue("N/A");

				panel.getChildren().add(text);
				i++;
			}
		}
	}

	private void clearDashboard() {
		// Run through all columns of the dashboard and remove the items
		final List<DashboardColumn> columns = this.dashboard.getModel().getColumns();
		for (final DashboardColumn column : columns) {
			column.getWidgets().clear();
		}
		// Now run through the dashboard itself and remove the items as well
		this.dashboard.getChildren().clear();
	}

	/**
	 * Getter for the attribute {@link CurrentCockpitEditorBean#projectsBean}.
	 * 
	 * @return Delivers the value of the attribute.
	 */
	public ProjectsBean getProjectsBean() {
		return this.projectsBean;
	}

	/**
	 * Setter for the attribute {@link CurrentCockpitEditorBean#projectsBean}.
	 * 
	 * @param projectsBean
	 *            The new value of the attribute.
	 */
	public void setProjectsBean(final ProjectsBean projectsBean) {
		this.projectsBean = projectsBean;
	}

	/**
	 * This method initializes the bean by using the current project name to load the project. <b>Do not call this method manually. It will only be accessed by
	 * Spring.</b>
	 */
	public void initalize() {
		synchronized (this) {
			try {
				// Make sure that the initialization will only be done for the init request.
				if (!FacesContext.getCurrentInstance().isPostback()) {
					// Remember the given parameters
					this.project = this.projectsBean.openProject(this.projectName);
					if (this.project != null) {
						// Remember the current time! This is important for the later comparison of the time stamps.
						this.resetTimeStamp();
						// Update the class loader and the specific classes used within various methods in this bean
						this.reloadClassLoader();
						this.reloadClassesAndMethods();
						this.fillDashboard();
					}
				}
			} catch (final ProjectLoadException ex) {
				CurrentCockpitEditorBean.LOG.error("An error occured while loading the project.", ex);
				GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occured while loading the project.");
			}
		}
	}

	/**
	 * This method reloads the field containing the methods and classes, using the current class loader.
	 * 
	 * @throws ProjectLoadException
	 *             If one or more of the classes and methods necessary for {@link ClassAndMethodContainer} could not be found using the given class loader.
	 */
	private void reloadClassesAndMethods() throws ProjectLoadException {
		synchronized (this) {
			this.classAndMethodContainer = new ClassAndMethodContainer(this.classLoader);
		}
	}

	/**
	 * This method reloads the class loader. In other words: The class loader will always be able to load classes from the jar-files within the lib-folder of the
	 * project.
	 * 
	 * @throws ProjectLoadException
	 *             If something went wrong.
	 */
	private void reloadClassLoader() throws ProjectLoadException {
		synchronized (this) {
			try {
				this.classLoader = this.projectManagerFacade.getClassLoader(this.projectName); // NOPMD (ClassLoader)
			} catch (final NullPointerException ex) {
				throw new ProjectLoadException("Invalid class loader.", ex);
			} catch (final IOException ex) {
				throw new ProjectLoadException("Could not load classes.", ex);
			} catch (final ProjectNotExistingException ex) {
				throw new ProjectLoadException("Project does not exist.", ex);
			}
		}
	}

	/**
	 * This method delivers the project stored in this bean.
	 * 
	 * @return The project for this user.
	 */
	public MIProject getProject() {
		synchronized (this) {
			return this.project;
		}
	}

	/**
	 * This method sets the project stored within this bean and returns the new page for the navigation.
	 * 
	 * @param newName
	 *            The name of the project.
	 */
	public void setProjectName(final String newName) {
		synchronized (this) {
			this.projectName = newName;
		}
	}

	/**
	 * This method delivers the project name stored in this bean.
	 * 
	 * @return The project name for this user.
	 */
	public String getProjectName() {
		synchronized (this) {
			return this.projectName;
		}
	}

	/**
	 * This method can be used to get the description of a {@link Display}. The description is read via the annotation using the java reflection API.
	 * 
	 * @param display
	 *            The display whose description should be extracted.
	 * @return The description for the display or a substitute if none is available. This is in either case human readable.
	 */
	public String getDescription(final MIDisplay display) {
		try {
			final String classname = display.getParent().getClassname();
			final String displayName = display.getName();
			// Try to get the class
			final Class<?> clazz = this.classLoader.loadClass(classname);
			// Run through all methods of the class
			for (final Method method : clazz.getMethods()) {
				// Get the potential annotation
				final Annotation annotationDisplay = method.getAnnotation(this.classAndMethodContainer.getDisplayAnnotationClass());
				if (annotationDisplay != null) {
					final String potDisplayName = (String) ClassAndMethodContainer.invokeMethod(this.classAndMethodContainer.getDisplayNameMethod(),
							annotationDisplay,
							"");
					if (potDisplayName.equals(displayName)) {
						return (String) ClassAndMethodContainer.invokeMethod(this.classAndMethodContainer.getDisplayDescriptionMethod(), annotationDisplay,
								"No description available");
					}
				}
			}
		} catch (final ClassNotFoundException ex) {
			CurrentCockpitEditorBean.LOG.warn("Class not found.", ex);
		}
		return "No description available";
	}

	/**
	 * This method tries to save the current project and informs the user about success or fail.
	 * 
	 * @param overwriteNewerProject
	 *            This flag determines whether a newer project should be overwritten.
	 */
	public void saveProject(final boolean overwriteNewerProject) {
		synchronized (this) {
			try {
				this.projectManagerFacade.saveProject(this.projectName, this.project, this.timeStamp, overwriteNewerProject);
				GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_INFO, this.globalPropertiesBean.getMsgProjectSaved());
				// Update the time stamp!
				this.resetTimeStamp();
			} catch (final IOException ex) {
				CurrentCockpitEditorBean.LOG.error("An error occured while saving the projet.", ex);
				GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_ERROR, "An error occured while saving the project.");
			} catch (final NewerProjectException ex) {
				CurrentCockpitEditorBean.LOG.info("The project has been modified externally in the meanwhile.", ex);
				GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_WARN, "The project has been modified externally in the meanwhile.");
				// Give the user the possibility to force-save the project
				RequestContext.getCurrentInstance().execute("forceSaveDlg.show()");
			} catch (final ProjectNotExistingException ex) {
				CurrentCockpitEditorBean.LOG.error("A project with the given name does not exist.", ex);
				GlobalPropertiesBean.showMessage(FacesMessage.SEVERITY_ERROR, "A project with the given name does not exist.");
			}
		}
	}

	/**
	 * This method sets the time stamp to the current system time.
	 */
	public void resetTimeStamp() {
		synchronized (this) {
			this.timeStamp = System.currentTimeMillis();
		}
	}

	/**
	 * This method delivers the available displays of the given plugin.
	 * 
	 * @param clazz
	 *            The plugin whose displays should be delivered.
	 * @return The available displays.
	 */
	public List<Display> getDisplays(final Class<AbstractPlugin> clazz) {
		final List<Display> result = new ArrayList<Display>();

		synchronized (this) {
			if (clazz != null) {
				// Run through all available methods and find the annotated ones
				final Method[] methods = clazz.getMethods();
				for (final Method method : methods) {
					final Display displayAnnot = method.getAnnotation(Display.class);
					// Check whether the annotation is available. If yes - add it to our result list
					if (displayAnnot != null) {
						result.add(displayAnnot);
					}
				}
			}
		}

		return result;
	}

	/**
	 * This method adds a new view to the project.
	 * 
	 * @param viewName
	 *            The name of the new view.
	 */
	public void addView(final String viewName) {
		synchronized (this) {
			if (this.project != null) {
				// Create the view and add it to our project
				final MIView view = this.factory.createView();
				view.setName(viewName);
				view.setDescription("No description available.");
				this.project.getViews().add(view);
			}
		}
	}

	/**
	 * Delivers the currently selected view.
	 * 
	 * @return The currently active view.
	 */
	public MIView getActiveView() {
		synchronized (this) {
			return this.activeView;
		}
	}

	/**
	 * This method sets the active view to a new value.
	 * 
	 * @param view
	 *            The new active view.
	 */
	public void setActiveView(final MIView view) {
		synchronized (this) {
			this.activeView = view;

			this.fillDashboard();
		}
	}

	/**
	 * This method adds the given display to the currently active view. If no view exists, this method does nothing.
	 * 
	 * @param display
	 *            The display which should be added to the current view.
	 */
	public void addDisplayToView(final MIDisplay display) {
		synchronized (this) {
			if (this.activeView != null) {
				final MIDisplayConnector connector = this.factory.createDisplayConnector();
				connector.setDisplay(display);
				connector.setName(UUID.randomUUID().toString());
				this.activeView.getDisplayConnectors().add(connector);

				// Now add it to the dashboard as well
				final FacesContext fc = FacesContext.getCurrentInstance();
				final Application application = fc.getApplication();

				final Panel panel = (Panel) application.createComponent(fc, "org.primefaces.component.Panel", "org.primefaces.component.PanelRenderer");
				panel.setId("displayConnectorPanel_" + (this.dashboard.getChildCount() + 1));
				panel.setHeader(connector.getName());
				panel.setClosable(true);
				panel.setToggleable(false);

				this.getDashboard().getChildren().add(panel);
				final DashboardColumn column = this.dashboardModel.getColumn(CurrentCockpitEditorBean.NUMBER_COLUMNS - 1);
				column.addWidget(panel.getId());
				final HtmlOutputText text = new HtmlOutputText();
				text.setValue("N/A");

				panel.getChildren().add(text);
			}
		}
	}

	/**
	 * This event is used if the currently selected tab changes. It modifies the active view property within this bean.
	 * 
	 * @param event
	 *            The onChange-event.
	 */
	public void onChange(final TabChangeEvent event) {
		if (event.getData() instanceof MIView) {
			this.setActiveView((MIView) event.getData());
		}
	}

	/**
	 * This method checks whether a display connector with the given name exists already.
	 * 
	 * @param name
	 *            The name to be checked.
	 * @return true iff the name exists already.
	 */
	private boolean existsDisplayConnectorName(final String name) {
		synchronized (this) {
			// Make sure a view is selected
			if (this.activeView == null) {
				return false;
			}

			// Run through all display connectors and check the name against the given one
			for (final MIDisplayConnector connector : this.activeView.getDisplayConnectors()) {
				if (connector.getName().equals(name)) {
					return true;
				}
			}

			// The name has not been found
			return false;
		}
	}

	/**
	 * This method is used as a validator for new display connector names.
	 * 
	 * @param context
	 *            The context of the validation.
	 * @param toValidate
	 *            The components which has be validated.
	 * @param value
	 *            The new value.
	 */
	public void validateDisplayConnectorName(final FacesContext context, final UIComponent toValidate, final Object value) {
		synchronized (this) {
			if ((value instanceof String) && (toValidate instanceof UIInput)) {
				final boolean nameExists = this.existsDisplayConnectorName((String) value);
				((UIInput) toValidate).setValid(!nameExists);
			}
		}
	}

	public Dashboard getDashboard() {
		return this.dashboard;
	}

	public void setDashboard(final Dashboard dashboard) {
		this.dashboard = dashboard;
		this.dashboard.setModel(this.dashboardModel);
	}

	/**
	 * The getter for the property {@link CurrentCockpitEditorBean#globalPropertiesBean}.
	 * 
	 * @return The current value of the property.
	 */
	public GlobalPropertiesBean getGlobalPropertiesBean() {
		return this.globalPropertiesBean;
	}

	/**
	 * The setter for the property {@link CurrentCockpitEditorBean#globalPropertiesBean}.
	 * 
	 * @param globalPropertiesBean
	 *            The new value of the property.
	 */
	public void setGlobalPropertiesBean(final GlobalPropertiesBean globalPropertiesBean) {
		this.globalPropertiesBean = globalPropertiesBean;
	}

}
